import argparse
import os
import sys
import time

import txaio
import yaml
from ocs import ocs_agent, site_config
from ocs.ocs_twisted import Pacemaker, TimeoutLock
from pyModbusTCP.client import ModbusClient

byteorder = sys.byteorder


def load_configs(dir_name, config_extension='yaml'):
    '''Loads all register configuration files form the specified directory (path).
    The configurations are returned as a list of the configurations from individual files.'''

    config_dir = os.environ.get('OCS_CONFIG_DIR')
    path = os.path.join(config_dir, 'generator_config', dir_name)

    all_configs = []

    ls = os.listdir(path)

    # Make a filter so that we only try to load files ending in config_extension
    def filt(f_name):
        return f_name.endswith(config_extension)
    # Filter
    configs = filter(filt, ls)
    # Load configurations from the remaining files
    for f in configs:
        read_config = os.path.join(path, f)
        with open(read_config) as f:
            try:
                data = yaml.load(f, Loader=yaml.SafeLoader)
            except BaseException:
                pass
            all_configs.append(data)
    return all_configs


def twos(val, bytes):
    '''Take an unsigned integer representation of a two's compilment integer and return the correctly signed integer'''
    b = val.to_bytes(bytes, byteorder=byteorder, signed=False)
    return int.from_bytes(b, byteorder=byteorder, signed=True)


def interp_unsigned_double_reg(r1, r2):
    '''Take two 16 bit register values and combine them assuming we really wanted to read a 32 bit unsigned register'''
    return (r1 << 16) + r2


def interp_signed_double_reg(r1, r2):
    '''Take two 16 bit register values and combine them assuming we really wanted to read a 32 bit signed register'''
    return twos(interp_unsigned_double_reg(r1, r2), 4)


def make_bin_reader(offset, spec):
    '''Read an individual bit or continuous range of bits out of the 2 byte register.
    A single bit can be specified as a single number between 1 and 16, and a range can
    be specified with a dash, i.e. X-Y where both X and Y are in the range
    1 to 16 and X < Y. In either case the bit or range of bits will be returned as an
    integer.'''
    spec = spec.split(' ')[1:]
    spec = spec[0].split('-')
    spec = [int(s) for s in spec]
    if len(spec) == 1:
        # Process individual bit
        spec = spec[0]
        # The mask leaves only the desired bit
        mask = sum([1 << s for s in range(spec - 1, spec)])

        def reader(val):
            return (val[offset] & mask) >> spec - 1
        return reader
    elif len(spec) == 2:
        # Process range
        low = spec[0]
        high = spec[1]
        if low >= high:
            raise ValueError('First bit in range specification must be smaller than last.')
        # The mask leaves only the desired bits
        mask = sum([1 << s for s in range(low - 1, high)])

        def reader(val):
            return (val[offset] & mask) >> low - 1
        return reader
    else:
        raise ValueError('Cannot read binary read_as specification; use single bit or continuous range.')


class ReadBlock(object):
    '''An object for reading, converting, and evaluating information from a single contiouous block of registers'''

    def __init__(self, config, error_out_of_range=True, filter_errors=True):
        self.name = config['block_name']

        try:
            self.read_start = config['read_start']
        except KeyError:
            self.read_start = config['page'] * 256

        self.read_len = config['read_len']
        self.functions = []
        self.rconfig = config['registers']
        self.error_val = None
        self.error_out_of_range = error_out_of_range
        self.filter_errors = filter_errors
        for i in self.rconfig:
            self.functions.append(self.build_reader_function(i, self.rconfig[i]))

    def build_reader_function(self, name, rconfig):
        '''Build and return a closure around a function that converts a specified piece of information from the block'''
        offset = rconfig['offset']

        if rconfig['read_as'] == '16U':
            def evaluator(registers):
                return registers[offset]
        elif rconfig['read_as'] == '16S':
            def evaluator(registers):
                return twos(registers[offset], 2)
        elif rconfig['read_as'] == '32U':
            def evaluator(registers):
                return interp_unsigned_double_reg(registers[offset], registers[offset + 1])
        elif rconfig['read_as'] == '32S':
            def evaluator(registers):
                return interp_signed_double_reg(registers[offset], registers[offset + 1])
        elif 'bin' in rconfig['read_as']:
            evaluator = make_bin_reader(offset, rconfig['read_as'])
        else:
            def evaluator(registers):
                return self.error_val

        def process(registers):
            val = evaluator(registers)
            if 'scale' in rconfig:
                val = val * rconfig['scale']

            if self.error_out_of_range:
                try:
                    if val < rconfig['min_val']:
                        val = self.error_val
                except KeyError:
                    pass

                try:
                    if val > rconfig['max_val']:
                        val = self.error_val
                except KeyError:
                    pass

            if val != self.error_val or not self.filter_errors:
                return {name: {'value': float(val), 'units': rconfig['units']}}
            else:
                return None

        return process

    def read(self, client):
        # Perform the read for the entire block
        registers = client.read_holding_registers(self.read_start, self.read_len)
        return_data = {}
        try:
            # Iterate through the functions that convert and return the individual pieces of
            # data from this block
            for f in self.functions:
                this_data = f(registers)
                if this_data is not None:
                    return_data.update(this_data)
        except Exception as e:
            # print(registers)
            print(f'Error in processing data: {e}')

        return return_data


class Generator:
    """Functions to communite with the Generator controller

    Parameters
    ----------
    host : str
        Address of the generator controller.
    port : int
        Port to generator controller, default to 5021.
    config_dir : string
        Sub-directory of .yaml configuration files that specify blocks of registers to read
        and specifies how to convert them into useable data.
    block_space_time : float
        Amount of time (in seconds) to wait between issuing seperate read_multiple_registers
        commands to the device. DSE device appears to crash without this waiting period.
    close_port : boolean
       Whether or not to close the open port to the device while waiting for the next Pacemaker
       triggered read cycle. The idea here is that closing the port alllows DSEWebNet to function
       in parallel with the agent.

    Attributes
    ----------
    read_blocks : list
        List of ReadBlock objects that represent the different continuous register locations to
        read from as specified in the config files.
    client : ModbusClient
        ModbusClient object that initializes connection
    """

    def __init__(self, host, port, config_dir, block_space_time=.1, close_port=True):
        self.host = host
        self.port = port
        self.read_config = load_configs(config_dir)
        self._build_config()
        self.close_port = close_port
        self.block_space_time = block_space_time
        self.client = ModbusClient(self.host, self.port, auto_open=True, auto_close=False)
        self.client.open()

    def _build_config(self):
        self.read_blocks = []
        for block in self.read_config:
            self.read_blocks.append(ReadBlock(block))

    def read_cycle(self):
        this_cycle_data = {}
        for i, val in enumerate(self.read_blocks):
            data = self._read_regs(val)
            this_cycle_data.update(data)
            time.sleep(self.block_space_time)  # A gap in time is required between individual requests,
            # i.e. a pause between reading each read_multiple_registers command.
        return this_cycle_data

    def _read_regs(self, register_block_object):
        if self.close_port:
            self.client.open()
        try:
            data = register_block_object.read(self.client)
        except Exception as e:
            print('error in read', e)
            if self.close_port:
                self.client.open()
            return
        if self.close_port:
            self.client.open()
        return data


class GeneratorAgent:
    """Monitor the Generator controller via ModBus.

    Parameters
    ----------
    agent : OCSAgent
        OCSAgent object which forms this Agent
    configdir : str
        Directory where .yaml configuration files specific to this generator instance
        are stored.
    host : str
        Address of the generator controller.
    port : int
        Port to generator controller, default to 5021.
    sample_interval : float
        Time between samples in seconds.

    Attributes
    ----------
    agent : OCSAgent
        OCSAgent object which forms this Agent
    take_data : bool
        Tracks whether or not the agent is actively issuing SNMP GET commands
        to the ibootbar. Setting to false stops sending commands.
    log : txaio.tx.Logger
        txaio logger object, created by the OCSAgent
    """

    def __init__(self, agent, configdir, host='localhost', port=5021, sample_interval=10.):

        self.host = host
        self.port = port

        self.agent: ocs_agent.OCSAgent = agent
        self.log = agent.log
        self.lock = TimeoutLock()

        self.configdir = configdir

        self.pacemaker_freq = 1. / sample_interval

        self.initialized = False
        self.take_data = False

        self.generator = None

        agg_params = {
            'frame_length': 10 * 60  # [sec]
        }
        self.agent.register_feed('generator',
                                 record=True,
                                 agg_params=agg_params,
                                 buffer_time=0)

    def _connect(self):
        """connect()
        Instantiates Generator object and check if client is open
        """

        self.generator = Generator(self.host, self.port, config_dir=self.configdir)
        if self.generator.client.is_open:
            self.initialized = True
        else:
            self.initialized = False

    @ocs_agent.param('auto_acquire', default=False, type=bool)
    def init_generator(self, session, params=None):
        """init_generator(auto_acquire=False)

        **Task** - Perform first time setup of the Generator.

        Parameters:
            auto_acquire (bool, optional): Starts data acquisition after
                initialization if True. Defaults to False.

        """

        if self.initialized:
            return True, "Already initialized."

        with self.lock.acquire_timeout(3, job='init') as acquired:
            if not acquired:
                self.log.warn("Could not start init because "
                              "{} is already running".format(self.lock.job))
                return False, "Could not acquire lock."

            self._connect()
            if not self.initialized:
                return False, 'Could not connect to generator'

        # Start data acquisition if requested
        if params['auto_acquire']:
            self.agent.start('acq')

        return True, 'Generator initialized.'

    @ocs_agent.param('_')
    def acq(self, session, params=None):
        """acq()

        **Process** - Start data acquisition.

        Notes
        -----
        The most recent data collected is stored in session data in the
        structure::

            >>> response.session['data']
            {"fields":
                {'Oil_pressure': {'value': 100.0, 'units': 'Kpa'},
                 'Coolant_temperature': {'value': 20.0, 'units': 'Degrees C'},
                 'Oil_temperature': {'value': 20.0, 'units': 'Degrees C'},
                 'Fuel_level': {'value': 100.0, 'units': '%'},
                 'Charge_alternator_voltage': {'value': 10.0, 'units': 'V'},
                 'Engine_Battery_voltage': {'value': 10.0, 'units': 'V'},
                 'Engine_speed': {'value': 4000, 'units': 'RPM'},
                 'Generator_frequency': {'value': 1.0, 'units': 'Hz'},
                 ...
                 'connection': {'last_attempt': 1680812613.939653, 'connected': True}},
             "address": 'localhost',
             "timestamp":1601925677.6914878}


        """

        with self.lock.acquire_timeout(0, job='acq') as acquired:
            if not acquired:
                self.log.warn("Could not start acq because {} is already running"
                              .format(self.lock.job))
                return False, "Could not acquire lock."

            self.take_data = True

            session.data = {"fields": {}}

            pm = Pacemaker(self.pacemaker_freq)
            while self.take_data:
                pm.sleep()

                current_time = time.time()
                data = {
                    'timestamp': current_time,
                    'connection': {},
                    'block_name': 'registers',
                    'data': {}
                }
                if not self.generator.client.is_open:
                    self.initialized = False

                # Try to re-initialize if connection lost
                if not self.initialized:
                    self._connect()

                # Only get readings if connected
                if self.initialized:
                    session.data.update({'connection': {'last_attempt': time.time(),
                                                        'connected': True}})
                    session.data['address'] = self.host

                    regdata = self.generator.read_cycle()
                    if regdata:
                        for reg in regdata:
                            data['data'][reg] = regdata[reg]["value"]
                            field_dict = {reg: regdata[reg]['value']}
                            session.data['fields'].update(field_dict)
                        session.data.update({'timestamp': current_time})
                    else:
                        self.log.info('Connection error or error in processing data.')
                        self.initialized = False

                # Continue trying to connect
                if not self.initialized:
                    session.data.update({'connection': {'last_attempt': time.time(),
                                                        'connected': False}})
                    self.log.info('Trying to reconnect.')
                    continue

                for field, val in data['data'].items():
                    _data = {
                        'timestamp': current_time,
                        'block_name': field,
                        'data': {field: val}
                    }
                    self.agent.publish_to_feed('generator', _data)

            self.agent.feeds['generator'].flush_buffer()

        return True, 'Acquisition exited cleanly.'

    def _stop_acq(self, session, params=None):
        """
        Stops acq process.
        """
        if self.take_data:
            self.take_data = False
            return True, 'requested to stop taking data.'
        else:
            return False, 'acq is not currently running'


def make_parser(parser=None):
    if parser is None:
        parser = argparse.ArgumentParser()

    pgroup = parser.add_argument_group('Agent Options')
    pgroup.add_argument("--host", help="Address to listen to.")
    pgroup.add_argument("--port", default=5021,
                        help="Port to listen on.")
    pgroup.add_argument('--mode', type=str, choices=['idle', 'init', 'acq'],
                        help="Starting action for the agent.")
    pgroup.add_argument("--configdir", type=str, help="Path to directory containing .yaml config files.")
    pgroup.add_argument("--sample-interval", type=float, default=10., help="Time between samples in seconds.")

    return parser


def main(args=None):
    # Start logging
    txaio.start_logging(level=os.environ.get("LOGLEVEL", "info"))

    parser = make_parser()

    # Interpret options in the context of site_config.
    args = site_config.parse_args(agent_class='GeneratorAgent',
                                  parser=parser,
                                  args=args)

    # Automatically acquire data if requested (default)
    init_params = False
    if args.mode == 'init':
        init_params = {'auto_acquire': False}
    elif args.mode == 'acq':
        init_params = {'auto_acquire': True}

    agent, runner = ocs_agent.init_site_agent(args)

    p = GeneratorAgent(agent,
                       configdir=args.configdir,
                       host=args.host,
                       port=int(args.port),
                       sample_interval=args.sample_interval)

    agent.register_task('init_generator', p.init_generator,
                        startup=init_params)
    agent.register_process('acq', p.acq, p._stop_acq)

    runner.run(agent, auto_reconnect=True)


if __name__ == '__main__':
    main()



import argparse
import socket
import time
from os import environ

import txaio
from ocs import ocs_agent, site_config
from ocs.ocs_twisted import TimeoutLock

from socs.agents.srs_cg635.drivers import SRSCG635Interface


class SRSCG635Agent:
    """Class to retrieve data from the SRS CG635 clock.

    Parameters
    ----------
    agent : ocs_agent.OCSAgent
        Instantiated OCSAgent class for this Agent.
    ip_address : str
        IP address of the Prologix GPIB interface.
    gpib_slot : int
        GPIB address set on the SRS CG635.
    """

    def __init__(self, agent, ip_address, gpib_slot):
        self.agent = agent
        self.log = agent.log
        self.lock = TimeoutLock()

        self.job = None
        self.ip_address = ip_address
        self.gpib_slot = gpib_slot
        self.monitor = False

        self.clock = None

        agg_params = {
            'frame_length': 10 * 60,
        }
        self.agent.register_feed('srs_clock',
                                 record=True,
                                 agg_params=agg_params,
                                 buffer_time=0)

    @ocs_agent.param('auto_acquire', default=False, type=bool)
    def init(self, session, params=None):
        """init(auto_acquire=False)

        **Task** - Initialize the connection to the srs clock.

        Parameters
        ----------
        auto_acquire: bool, optional
            Default is False. Starts data acquisition after initialization
            if True.
        """
        with self.lock.acquire_timeout(0) as acquired:
            if not acquired:
                return False, "Could not acquire lock"

            try:
                self.clock = SRSCG635Interface(self.ip_address, self.gpib_slot)
                self.idn = self.clock.identify()

            except (socket.timeout, OSError) as e:
                self.log.error(f"Clock timed out during connect: {e}")
                return False, "Timeout"
            self.log.info("Connected to Clock: {}".format(self.idn))

        # Start data acquisition if requested in site-config
        if params['auto_acquire']:
            self.agent.start('acq')

        return True, 'Initialized Clock.'

    def _initialize_interface(self):
        """Initialize the SRSCG635 interface."""
        try:
            self.clock = SRSCG635Interface(self.ip_address, self.gpib_slot)
        except (socket.timeout, OSError) as e:
            self.log.error(f"Clock timed out during connect: {e}")
            self.clock = None
            raise ConnectionError

        self.idn = self.clock.identify()
        self.log.info("Connected to Clock: {}".format(self.idn))
        self.log.info("Clearing event registers and error queue.")
        self.clock.clear()

    @ocs_agent.param('test_mode', default=False, type=bool)
    @ocs_agent.param('wait', default=1, type=float)
    def acq(self, session, params):
        """acq(wait=1, test_mode=False)

        **Process** - Continuously monitor SRS clock registers
        and publish to a feed. Note that local changes on the clock
        (physically pressing buttons) will be unavailable during this
        process.

        The ``session.data`` object stores the most recent published values
        in a dictionary. For example::

            session.data = {
                'timestamp': 1598626144.5365012,
                'block_name': 'clock_output',
                'data': {
                    'Frequency': 122880000.0
                    'Standard_CMOS_Output': 3,
                    'Running_State': 1,
                    'Timebase': 3
                }
            }

        Refer to drivers for interpretation of outputs.

        Parameters
        ----------
        wait : float, optional
            Time to wait between measurements [seconds]. Default=1s.
        test_mode : bool, optional
            Run the process loop only once. Meant for testing only.
            Default=False

        """
        self.monitor = True

        while self.monitor:
            with self.lock.acquire_timeout(1) as acquired:
                if acquired:
                    if not self.clock:
                        self.log.info("Trying to reconnect...")
                        try:
                            self._initialize_interface()
                            if session.degraded:
                                self.log.info("Connection re-established.")
                                session.degraded = False
                        except ConnectionError:
                            self.log.error("Failed to reconnect. Check network connection.")
                            session.degraded = True
                            time.sleep(1)
                            continue

                    data = {
                        'timestamp': time.time(),
                        'block_name': 'clock_output',
                        'data': {}
                    }

                    try:
                        freq, stdc, runs, timb = self.clock.get_all_status()
                        data['data']['Frequency'] = freq
                        data['data']['Standard_CMOS_Output'] = stdc
                        data['data']['Running_State'] = runs
                        data['data']['Timebase'] = timb

                    except socket.timeout as e:
                        self.log.error(f"Timeout in retrieving clock data: {e}")
                        self.clock = None
                        session.degraded = True
                        continue

                    self.agent.publish_to_feed('srs_clock', data)

                    # Allow this process to be queried to return current data
                    session.data = data

                else:
                    self.log.warn("Could not acquire in monitor clock")

            time.sleep(params['wait'])

            if params['test_mode']:
                break

        return True, "Finished monitoring clock"

    def _stop_acq(self, session, params):
        """Stop monitoring the clock output."""
        if self.monitor:
            self.monitor = False
            return True, 'requested to stop taking data.'
        else:
            return False, 'acq is not currently running'


def make_parser(parser=None):
    """Build the argument parser for the Agent. Allows sphinx to automatically
    build documentation based on this function.

    """
    if parser is None:
        parser = argparse.ArgumentParser()

    # Add options specific to this agent.
    pgroup = parser.add_argument_group('Agent Options')
    pgroup.add_argument('--ip-address', type=str, help="Prologix IP Address")
    pgroup.add_argument('--gpib-slot', type=int, help="Internal SRS GPIB Address")
    pgroup.add_argument('--mode', type=str, default='acq',
                        choices=['idle', 'init', 'acq'],
                        help="Starting action for the Agent.")

    return parser


def main(args=None):
    # Start logging
    txaio.start_logging(level=environ.get("LOGLEVEL", "info"))

    parser = make_parser()
    args = site_config.parse_args(agent_class='SRSCG635Agent',
                                  parser=parser,
                                  args=args)

    if args.mode == 'init':
        init_params = {'auto_acquire': False}
    if args.mode == 'acq':
        init_params = {'auto_acquire': True}

    agent, runner = ocs_agent.init_site_agent(args)

    p = SRSCG635Agent(agent, args.ip_address, args.gpib_slot)

    agent.register_task('init', p.init, startup=init_params)
    agent.register_process('acq', p.acq, p._stop_acq)

    runner.run(agent, auto_reconnect=True)


if __name__ == '__main__':
    main()



import socket
import time

protection_status_key = [
    'Over voltage',
    'Over current',
    'AC power failure or power interuption',
    '',
    'Over temperature',
    '',
    'IOC communication error',
    '',
]


class PMX:
    """The PMX object for communicating with the Kikusui PMX power supplies.
    Args:
        tcp_ip (str): TCP IP address
        tcp_port (int): TCP port
    """

    def __init__(self, ip, port):
        self.ip = ip
        self.port = port
        self.wait_time = 0.01
        self.buffer_size = 128
        self.conn = self._establish_connection(self.ip, int(self.port))

    def _establish_connection(self, ip, port, timeout=2):
        conn = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        conn.settimeout(timeout)
        attempts = 3
        for attempt in range(attempts):
            try:
                conn.connect((ip, port))
                break
            except (ConnectionRefusedError, OSError):
                print(f"Failed to connect to device at {ip}:{port}")
        else:
            raise RuntimeError('Could not connect to PMX.')
        return conn

    def close(self):
        self.conn.close()

    def send_message(self, msg, read=True):
        for attempt in range(2):
            try:
                self.conn.sendall(msg)
                time.sleep(0.5)
                if read:
                    data = self.conn.recv(self.buffer_size).strip().decode('utf-8')
                    return data
                return
            except (socket.timeout, OSError):
                print("Caught timeout waiting for responce from PMX. Trying again...")
                time.sleep(1)
                if attempt == 1:
                    print("Resetting connection")
                    self.conn.close()
                    self.conn = self._establish_connection(self.ip, int(self.port))
                    return self.send_message(msg, read=read)

    def wait(self):
        time.sleep(self.wait_time)

    def check_output(self):
        """ Return the output status """
        val = int(self.send_message(b'output?\n'))
        msg = "Measured output state = "
        states = {0: 'OFF', 1: 'ON'}
        if val in states:
            msg += states[val]
        else:
            msg += 'Fail'
        return msg, val

    def check_error(self):
        """ Check oldest error from error queues. Error queues store up to 255 errors """
        val = self.send_message(b':system:error?\n')
        code, msg = val.split(',')
        code = int(code)
        msg = msg[1:-2]
        return msg, code

    def clear_alarm(self):
        """ Clear alarm """
        self.send_message(b'output:protection:clear\n', read=False)

    def turn_on(self):
        """ Turn the PMX on """
        self.send_message(b'output 1\n', read=False)
        self.wait()
        return self.check_output()

    def turn_off(self):
        """ Turn the PMX off """
        self.send_message(b'output 0\n', read=False)
        self.wait()
        return self.check_output()

    def check_current(self):
        """ Check the current setting """
        val = float(self.send_message(b'curr?\n'))
        msg = "Current setting = {:.3f} A".format(val)
        return msg, val

    def check_voltage(self):
        """ Check the voltage setting """
        val = float(self.send_message(b'volt?\n'))
        msg = "Voltage setting = {:.3f} V".format(val)
        return msg, val

    def meas_current(self):
        """ Measure the current """
        val = float(self.send_message(b'meas:curr?\n'))
        msg = "Measured current = {:.3f} A".format(val)
        return msg, val

    def meas_voltage(self):
        """ Measure the voltage """
        val = float(self.send_message(b'meas:volt?\n'))
        msg = "Measured voltage = {:.3f} V".format(val)
        return msg, val

    def set_current(self, curr):
        """ Set the current """
        self.send_message(b'curr %a\n' % curr, read=False)
        self.wait()
        return self.check_current()

    def set_voltage(self, vol):
        """ Set the voltage """
        self.send_message(b'volt %a\n' % vol, read=False)
        self.wait()
        return self.check_voltage()

    def check_source(self):
        """ Check the source of PMX """
        val = self.send_message(b'volt:ext:sour?\n')
        msg = "Source: " + val
        return msg, val

    def use_external_voltage(self):
        """ Set PMX to use external voltage """
        self.send_message(b'volt:ext:sour volt\n', read=False)
        self.wait()
        return self.check_source()

    def ign_external_voltage(self):
        """ Set PMX to ignore external voltage """
        self.send_message(b'volt:ext:sour none\n', read=False)
        self.wait()
        return self.check_source()

    def check_current_limit(self):
        """ Check the PMX current protection limit """
        val = float(self.send_message(b'curr:prot?\n'))
        msg = "Current protection limit = {:.3f} A".format(val)
        return msg, val

    def check_voltage_limit(self):
        """ Check the PMX voltage protection limit """
        val = float(self.send_message(b'volt:prot?\n'))
        msg = "Voltage protection limit = {:.3f} V".format(val)
        return msg, val

    def set_current_limit(self, curr_lim):
        """ Set the PMX current protection limit """
        self.send_message(b'curr:prot %a\n' % curr_lim, read=False)
        self.wait()
        return self.check_current_limit()

    def set_voltage_limit(self, vol_lim):
        """ Set the PMX voltage protection limit """
        self.send_message(b'volt:prot %a\n' % vol_lim, read=False)
        self.wait()
        return self.check_voltage_limit()

    def check_prot(self):
        """ Check the protection status
        Return:
            val (int): protection status code
        """
        val = int(self.send_message(b'stat:ques?\n'))
        return val

    def get_prot_msg(self, val):
        """ Get the protection status message
        Args:
            val (int): protection status code
        Return:
            msg (str): protection status message
        """
        msg = []
        for i in range(8):
            if (val >> i & 1):
                msg.append(protection_status_key[i])
        msg = ', '.join(msg)
        return msg
